<?xml version="1.0" encoding="utf-8" ?>
<ContentPage xmlns="http://schemas.microsoft.com/dotnet/2021/maui"
             xmlns:x="http://schemas.microsoft.com/winfx/2009/xaml"
             x:Class="Microsoft.ML.OnnxRuntime.InferenceSample.Maui.MainPage">

    <StackLayout>
        <Frame BackgroundColor="#2196F3" Padding="24" CornerRadius="0">
            <Label Text="Inference Sample" HorizontalTextAlignment="Center" TextColor="White" FontSize="36"/>
        </Frame>
        <Label Text="ONNX Runtime" FontSize="Title" Padding="30,10,30,10"/>
        <Button x:Name="Start" Text="Run tests" Clicked="Start_Clicked" FontSize="Large" />
        <Label x:Name="OutputLabel" Text="Output" FontSize="Small"/>
    </StackLayout>

</ContentPage>
