// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#pragma once

#include "winrt_headers.h"

#include "core/providers/winml/winml_provider_factory.h"
#include "adapter/winml_adapter_c_api.h"

#include "UniqueOrtPtr.h"
