/** @type {import('@jest/types').Config.InitialOptions} */
module.exports = {
  rootDir: '..',
  testMatch: ['<rootDir>/test/**/*.test.js'],
  testTimeout: 240000,
  maxWorkers: 1,
  globalSetup: 'detox/runners/jest/globalSetup',
  globalTeardown: 'detox/runners/jest/globalTeardown',
  reporters: ['jest-junit'],
  testEnvironment: 'detox/runners/jest/testEnvironment',
  verbose: true,
};
