// Copyright (c) Microsoft Corporation. All rights reserved.
// Copyright (c) Huawei. All rights reserved.
// Licensed under the MIT License

#pragma once

#include <acl/acl.h>
#include <acl/acl_op_compiler.h>
#include <acl/ops/acl_cblas.h>
#include <ge/ge_ir_build.h>
#include <ge/ge_api_types.h>
#include <parser/onnx_parser.h>
