<maui:MauiWinUIApplication
    x:Class="Microsoft.ML.OnnxRuntime.InferenceSample.Maui.WinUI.App"
    xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"
    xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
    xmlns:maui="using:Microsoft.Maui"
    xmlns:local="using:Microsoft.ML.OnnxRuntime.InferenceSample.Maui.WinUI">

</maui:MauiWinUIApplication>
