// Copyright (c) 2023 Advanced Micro Devices, Inc. All rights reserved.
// Licensed under the MIT License.
#pragma once

#ifdef ONNXRUNTIME_VITISAI_EP_STUB
#include <gsl/gsl>
#else
#include <gsl/gsl>
#endif
