// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.
//
//  AppDelegate.h
//  macos_package_test
//

#import "AppDelegate.h"

@interface AppDelegate ()

@end

@implementation AppDelegate

- (void)applicationDidFinishLaunching:(NSNotification*)aNotification {
  // Insert code here to initialize your application
}

- (void)applicationWillTerminate:(NSNotification*)aNotification {
  // Insert code here to tear down your application
}

- (BOOL)applicationSupportsSecureRestorableState:(NSApplication*)app {
  return YES;
}

@end
