<?xml version="1.0" encoding="utf-8"?>
<Project ToolsVersion="4.0" xmlns="http://schemas.microsoft.com/developer/msbuild/2003">
  <PropertyGroup>
    <WindowsAIVersion>[PackageVersion]</WindowsAIVersion>
  </PropertyGroup>
</Project>
