// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#pragma once

namespace onnxruntime {
namespace cuda {
namespace variadic_elementwise_ops {
struct Sum {};
struct Min {};
struct Max {};
}  // namespace variadic_elementwise_ops
}  // namespace cuda
}  // namespace onnxruntime
