// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#pragma once

#pragma warning(push)
#pragma warning(disable : 26495)
#pragma warning(disable : 6011)

#pragma warning(disable : 26451)
#pragma warning(disable : 6387)

#include "cppwinrt_onnx.h"
#include "dx.h"
#include "FeatureValues.h"

#pragma warning(pop)
