// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.
//
//  AppDelegate.h
//  macos_package_test
//

#import <Cocoa/Cocoa.h>

@interface AppDelegate : NSObject <NSApplicationDelegate>

@end
