// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.
//
//  AppDelegate.h
//  macos_package_test
//

#import <Cocoa/Cocoa.h>

int main(int argc, const char* argv[]) {
  @autoreleasepool {
    // Setup code that might create autoreleased objects goes here.
  }
  return NSApplicationMain(argc, argv);
}
