<?xml version="1.0" encoding="UTF-8" ?>
<Shell
    x:Class="Microsoft.ML.OnnxRuntime.Tests.MAUI.AppShell"
    xmlns="http://schemas.microsoft.com/dotnet/2021/maui"
    xmlns:x="http://schemas.microsoft.com/winfx/2009/xaml"
    xmlns:local="clr-namespace:Microsoft.ML.OnnxRuntime.Tests.MAUI"
    Shell.FlyoutBehavior="Disabled"
    Title="Microsoft.ML.OnnxRuntime.Tests.MAUI">

    <ShellContent
        Title="Home"
        ContentTemplate="{DataTemplate local:MainPage}"
        Route="MainPage" />

</Shell>
