// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#include "js_kernel.h"

namespace onnxruntime {
namespace js {
}  // namespace js
}  // namespace onnxruntime
