// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

declare module 'onnxruntime-web' {
  export * from 'onnxruntime-common';
}

declare module 'onnxruntime-web/all' {
  export * from 'onnxruntime-web';
}

declare module 'onnxruntime-web/wasm' {
  export * from 'onnxruntime-web';
}

declare module 'onnxruntime-web/webgl' {
  export * from 'onnxruntime-web';
}

declare module 'onnxruntime-web/webgpu' {
  export * from 'onnxruntime-web';
}
