// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

import{AppRegistry} from 'react-native';
import App from './src/App';
import{name as appName} from './app.json';

// Register React Native example app
AppRegistry.registerComponent(appName, () => App);
