// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

import { createRequire } from 'module';
const require = createRequire(import.meta.url);

const testFunction = require('./common');

export default testFunction;
