// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#pragma once

#if defined(__GNUC__)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif

#include <webgpu/webgpu_cpp.h>

#if defined(__GNUC__)
#pragma GCC diagnostic pop
#endif
