# Don't edit this file manully.
# Run {{ python_file_name }} to generate it.

name: {{ ci_workflow_name }}
on:
  pull_request:
    paths:
      - docs/**
      - README.md
      - CONTRIBUTING.md
      - BUILD.md

jobs:
{%- for name in job_names %}
  job{{ loop.index }}:
    name: {{ name }}
    runs-on: ["self-hosted", "1ES.Pool=onnxruntime-github-Ubuntu2204-AMD-CPU"]
    steps:
      - run: 'echo "No build required, only documentation changed"'
{% endfor %}
