// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#pragma once

// DirectX
#include <d3d12.h>
#include <D3d11_4.h>
#include <dxgi1_5.h>
#include <dxgi1_6.h>
#include <windows.graphics.directx.direct3d11.h>
#include <windows.graphics.directx.direct3d11.interop.h>
#include <directxmath.h>
#include <directxpackedvector.h>
