// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#pragma once

namespace onnxruntime {
namespace cuda {
template <typename T>
KernelCreateInfo BuildKernelCreateInfo();
}
}  // namespace onnxruntime
