// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#pragma once

#include "core/providers/js/js_kernel.h"

namespace onnxruntime {
namespace js {

JSEP_KERNEL_IMPL(Range, Range);

}  // namespace js
}  // namespace onnxruntime
