// Copyright (c) Microsoft Corporation. All rights reserved.
// Copyright (c) 2020, NXP Semiconductor, Inc. All rights reserved.
// Licensed under the MIT License.

#pragma once

namespace onnxruntime {
namespace armnn_ep {
template <typename T>
KernelCreateInfo BuildKernelCreateInfo();
}
}  // namespace onnxruntime
