// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#ifndef MNISTDataHandler_h
#define MNISTDataHandler_h

#import <react/RCTBridgeModule.h>

@interface MNISTDataHandler : NSObject <RCTBridgeModule>
@end

#endif /* MNISTDataHandler_h */
